import re
from inspect import getsource
from typing import Callable, List, Type

from pydantic import BaseModel

from nortech_internal.derivers.values.physical_units_schema import PhysicalQuantity
from nortech_internal.derivers.values.schema import (
    DeriverSchema,
    DeriverSchemaConfiguration,
    DeriverSchemaDAG,
    DeriverSchemaInput,
    DeriverSchemaOutput,
    DeriverSchemaOutputWithDAG,
    get_actual_type,
)


def input_schema_to_deriver_schema_inputs(
    input_schema: Type[BaseModel],
) -> List[DeriverSchemaInput]:
    input_fields = (
        (field_name, field)
        for field_name, field in input_schema.model_fields.items()
        if field_name != "timestamp"
    )

    return [
        DeriverSchemaInput(
            name=field_name,
            description=field.description,  # type: ignore
            dataType=get_actual_type(field.annotation),
            physicalQuantity=PhysicalQuantity(
                **field.json_schema_extra["physicalQuantity"]  # type: ignore
            ),  # type: ignore
            suggestedInputsFromOtherDerivers=[
                DeriverSchemaOutputWithDAG(
                    **suggestedInputFromOtherDeriver,
                    deriver_schema_DAG=get_deriver_schema_DAG(
                        suggestedInputFromOtherDeriver["create_deriver_schema"]
                    ),
                )
                for suggestedInputFromOtherDeriver in field.json_schema_extra[  # type: ignore
                    "suggestedInputsFromOtherDerivers"
                ]
            ],
        )
        for field_name, field in input_fields
    ]


def output_schema_to_deriver_schema_outputs(
    output_schema: Type[BaseModel],
) -> List[DeriverSchemaOutput]:
    output_fields = (
        (field_name, field)
        for field_name, field in output_schema.model_fields.items()
        if field_name != "timestamp"
    )

    return [
        DeriverSchemaOutput(
            name=field_name,
            description=field.description,  # type: ignore
            dataType=get_actual_type(field.annotation),
            physicalQuantity=PhysicalQuantity(
                **field.json_schema_extra["physicalQuantity"]  # type: ignore
            ),
        )
        for field_name, field in output_fields
    ]


def config_schema_to_config_dict(
    config_schema: Type[BaseModel],
) -> List[DeriverSchemaConfiguration]:
    return [
        DeriverSchemaConfiguration(
            name=field_name,
            description=field.description,  # type: ignore
            dataType=get_actual_type(field.annotation),
        )
        for field_name, field in config_schema.model_fields.items()
    ]


def get_deriver_schema_DAG(
    create_deriver_schema: Callable[[], DeriverSchema],
) -> DeriverSchemaDAG:
    script = getsource(create_deriver_schema)

    deriver_schema = create_deriver_schema()
    inputs = input_schema_to_deriver_schema_inputs(input_schema=deriver_schema.inputs)
    outputs = output_schema_to_deriver_schema_outputs(
        output_schema=deriver_schema.outputs,
    )
    configurations = config_schema_to_config_dict(
        config_schema=deriver_schema.configurations
    )

    return DeriverSchemaDAG(
        name=deriver_schema.name,
        description=deriver_schema.description,
        inputs=inputs,
        outputs=outputs,
        configurations=configurations,
        script=script,
    )


def get_create_deriver_schema(script: str) -> Callable[[], DeriverSchema]:  # type: ignore
    pattern = r"suggestedInputs=\[.*?\],?\s*"
    script = re.sub(pattern, "", script, flags=re.DOTALL)

    # Step 1: Prepare a local execution context
    local_exec_context = {}

    # Step 2: Execute the code within the local execution context
    exec(script, globals(), local_exec_context)

    # Step 3: Identify new functions in the local execution context
    new_functions = [obj for obj in local_exec_context.values() if callable(obj)]

    # Assuming there's only one new function for simplicity
    if new_functions:
        new_function = new_function = new_functions[
            0
        ]  # Assuming there's only one new function for simplicity
        # Step 4: Fix self-references by updating __globals__ with the identified function name
        # Find the name of the new function in the local_exec_context
        for name, obj in local_exec_context.items():
            if obj is new_function:
                new_function_name = name
                break

        # Update __globals__ for the new function to include a self-reference
        new_function.__globals__[new_function_name] = new_function  # type: ignore

        # Now you can call the new function, and it will be able to reference itself
        return new_function
